import cv2
import os
os.nice(20)
import subprocess
import numpy as np

def generate_plain_image(pixel_value=255, shape=(1000, 500), outfile_name="val_255.png"):
	"""Generates a plain image. Useful for figuring out whether a pixel value of 255 means white or 0 means white. (Hint: 255 is whie in png files)

	Args:
		pixel_value (int, optional): the pixel value to generate. Defaults to 255.
		shape (tuple, optional): size of image. Defaults to (1000, 500).
		outfile_name (str, optional): where to save the file. Defaults to "val_255.png".
	"""
	out = (np.ones(shape) * pixel_value).astype(np.uint8)
	cv2.imwrite(filename=outfile_name, img=out, )

if __name__ == '__main__':
	generate_plain_image()


import math
import time
from typing import Optional

from tqdm import tqdm as tqdm

"""
Build a super dumb iterator wrapper which then returns a progress, this should be kinda splitted
first we need an object which remains alive the whole time where the logic and the variables are set to know
where the progress should be displayed (in our case we set in the gui, that we want the progress bar to be updated)

The second part is the iterator which only needs to be alive during "work"

thats how an iterator is made: https://treyhunner.com/2018/06/how-to-make-an-iterator-in-python/



this should probably be in an extra package since tqdm is used in the unet_pytorch package and we should not create
circular dependencies between the packages


the persistent class could have a private class which is the iterator and a method for creating the iterator


"""


class ProgressNotifier:

    def __init__(self):
        self.__task_progress = None
        self.__task_progress_details = None
        self.__use_tqdm = False

    @staticmethod
    def progress_notifier_tqdm():
        # should create a progress-notifier object prepared for using tqdm as output
        notifier = ProgressNotifier()
        notifier.__use_tqdm = True
        return notifier

    @staticmethod
    def silent_notifier():
        """
        returns a progress notifier that is silent (no progress notifications)
        used for the case a method requires a progress notifier but we do not want to show anything
        """
        return ProgressNotifier()

    def iterator(self, iterable, total:Optional[int]=None):
        """
        iterable:
            element which can be iterated
        total: Optional[int] default None
            sets the total value, default is None.
        """
        # check if it is iterable
        try:
            iterator = iter(iterable)
            if self.__use_tqdm:
                return tqdm(iterable)
            return self.__IteratorWrapper(iterable, self.__task_progress, self.__task_progress_details,total)
        except TypeError:
            raise TypeError("object is not possible to iterate")

    def set_progress_report(self, task):
        # check if task meets the requirements
        try:
            task(0)
            self.__task_progress = task
        except:
            raise Exception('The given variable is not a function with 1 numeric parameter (or similar constructs)')
        # sets the task that should be called on progress change
        # task should be a method or lambda that takes a number as input

    def set_progress_detail(self, task):
        # check if task meets the requirements
        try:
            task(0, 0, 0, 0, 0, 0)
            self.__task_progress_details = task
        except:
            raise Exception('The given variable is not a function with 6 numeric parameters (or similar constructs)')

        # sets the task that should be called on progress change
        # it gives details about the progress (like tqdm does), ETA, iterations per second and so on
        # this var doesn't have to be set

    class __IteratorWrapper:
        # this is a class variable (in java like static)
        timeMultiplier = 1000  # time values in milli seconds

        def __init__(self, iterable, task_progress, task_progress_details=None,total=None):
            self.__total = None
            self.__time = None  # time used until "now", now is the current "next" call
            self.__eta = None  # calculated value of how long will it take

            self.__iterable = iterable
            self.__iterator = iter(iterable)
            self.__task_progress = task_progress
            self.__task_progress_details = task_progress_details
            self.__time_start = int(round(time.time() * self.timeMultiplier))
            self.__current = 0  # current count of iterations
            # init total iterations number if possible
            if total is None:
                if iterable is not None:
                    try:
                        self.__total = len(iterable)
                    except (TypeError, AttributeError):
                        self.__total = None
            else:
                self.__total=total
                pass
            if task_progress_details is not None:
                task_progress_details(0, 0, 0, 0, 0, 0)

        def __iter__(self):
            return self

        def __next__(self):
            self.__current += 1
            self.__time = int(round(time.time() * self.timeMultiplier))
            # eta is an approximation time until now divided by current and multiplied by total,
            # only if total is not None
            if self.__total is not None:
                self.__eta = ((self.__time - self.__time_start) / self.__current) * self.__total

            # report progress and return next element
            if self.__task_progress is not None:
                if self.__total is not None:  # if total is not None the current progress is reported
                    self.__task_progress(self.__current / self.__total)
                else:  # otherwise the number of iterations is reported
                    self.__task_progress(self.__current)

            # todo: the progress_details should update in an extra thread each second and the duration should be a timer in there
            # since there could be very slow iterations taking a while and with extra thread and duration as timer
            # the user "notices" progress
            if self.__task_progress_details is not None and self.__eta is not None:
                hh_eta = math.floor(self.__eta / (self.timeMultiplier * 3600))
                mm_eta = math.floor(self.__eta / (self.timeMultiplier * 60) - hh_eta * 60)
                ss_eta = math.floor((self.__eta / self.timeMultiplier) - (hh_eta * 3600 + mm_eta * 60))

                duration = (self.__time - self.__time_start) / self.timeMultiplier
                hh_current = math.floor(duration / 3600)
                mm_current = math.floor(duration / 60) - hh_current * 60
                ss_current = math.floor(duration) - (hh_current * 3600 + mm_current * 60)

                self.__task_progress_details(hh_current, mm_current, ss_current, hh_eta, mm_eta, ss_eta)

            return self.__iterator.__next__()



import os
import shutil
import time
import gc

import numpy as np
import tifffile

import biu.siam_unet as siam
import biu.unet as unet
from biu import unet3d
# create test folder with random training and test data
from biu.progress import ProgressNotifier

folder = './temp_test/'


def test_unet():
    folder_image = folder + 'training_data/image/'
    folder_mask = folder + 'training_data/mask/'
    folder_results = folder + 'results/'
    os.makedirs(folder_image, exist_ok=True)
    os.makedirs(folder_mask, exist_ok=True)
    os.makedirs(folder_results, exist_ok=True)

    for i in range(5):
        # regular unet
        random_image = np.random.randint(0, 255, (128, 128))
        random_mask = np.random.randint(0, 255, (128, 128))
        tifffile.imwrite(folder_image + f'{i}.tif', random_image)
        tifffile.imwrite(folder_mask + f'{i}.tif', random_mask)

    random_movie = np.random.randint(0, 255, (20, 128, 128))
    tifffile.imwrite(folder + 'movie.tif', random_movie)

    # create training data set
    data = unet.DataProcess(source_dir=(folder_image, folder_mask), dim_out=(64, 64), data_path=folder + 'data/')

    # train
    train = unet.Trainer(data, num_epochs=4, n_filter=8, save_dir=folder + 'models_unet/')
    train.start()

    # predict movie
    unet.Predict(folder + 'movie.tif', result_name=folder_results + 'movie.tif',
                 model_params=folder + 'models_unet/model.pth', resize_dim=(64, 64),
                 progress_notifier=ProgressNotifier())


def test_siam_unet():
    folder_image_siam = folder + 'training_data_siam/image/'
    folder_mask_siam = folder + 'training_data_siam/mask/'
    folder_data = folder + 'data/'
    folder_results = folder + 'results/'
    os.makedirs(folder, exist_ok=True)
    os.makedirs(folder_image_siam, exist_ok=True)
    os.makedirs(folder_mask_siam, exist_ok=True)
    os.makedirs(folder_data, exist_ok=True)
    os.makedirs(folder_results, exist_ok=True)

    for i in range(5):
        # siam unet
        random_image = np.random.randint(0, 255, (2, 128, 128))
        random_mask = np.random.randint(0, 255, (128, 128))
        tifffile.imwrite(folder_image_siam + f'{i}.tif', random_image)
        tifffile.imwrite(folder_mask_siam + f'{i}.tif', random_mask)

    random_movie = np.random.randint(0, 255, (20, 128, 128))
    tifffile.imwrite(folder + 'movie.tif', random_movie)

    # create training data set
    data_siam = siam.DataProcess(source_dir=(folder_image_siam, folder_mask_siam), dim_out=(64, 64),
                                 data_path=folder + 'data_siam/')

    # train
    train_siam = siam.Trainer(data_siam, num_epochs=4, n_filter=8, save_dir=folder + 'models_siam/', load_weights=None)
    train_siam.start()

    # predict movie
    siam.Predict(folder + 'movie.tif', result_name=folder_results + 'movie.tif',
                 model_params=folder + 'models_siam/model.pth', resize_dim=(64, 64),
                 progress_notifier=ProgressNotifier())

def test_unet3d():
    folder_image = folder + 'training_data_3d/image/'
    folder_mask = folder + 'training_data_3d/mask/'
    folder_results = folder + 'results/'
    os.makedirs(folder_image, exist_ok=True)
    os.makedirs(folder_mask, exist_ok=True)
    os.makedirs(folder_results, exist_ok=True)

    for i in range(5):
        # regular unet
        random_image = np.random.randint(0, 255, (32, 128, 128))
        random_mask = np.random.randint(0, 255, (32, 128, 128))
        tifffile.imwrite(folder_image + f'{i}.tif', random_image)
        tifffile.imwrite(folder_mask + f'{i}.tif', random_mask)

    random_movie = np.random.randint(0, 255, (32, 128, 128))
    tifffile.imwrite(folder + 'movie.tif', random_movie)

    # create training data set
    data = unet3d.DataProcess(source_dir=(folder_image, folder_mask), dim_out=(32, 64, 64), data_path=folder + 'data/')

    # train
    train = unet3d.Trainer(data, num_epochs=4, n_filter=8, save_dir=folder + 'models_unet3d/')
    train.start()

    # predict movie
    unet3d.Predict(folder + 'movie.tif', result_name=folder_results + 'movie.tif',
                 model_params=folder + 'models_unet3d/model.pth', resize_dim=(16, 64, 64),
                 progress_notifier=ProgressNotifier())


def delete_folder_with_retry(folder, max_attempts=5, wait_seconds=2):
    for attempt in range(max_attempts):
        try:
            gc.collect()  # Force garbage collection
            shutil.rmtree(folder)
            print(f"Successfully deleted {folder}")
            break
        except PermissionError as e:
            print(f"PermissionError on attempt {attempt + 1}: {e}")
            time.sleep(wait_seconds)
        except Exception as e:
            print(f"An error occurred: {e}")
            break
    else:
        print(f"Failed to delete {folder} after {max_attempts} attempts.")


if __name__ == "__main__":
    test_unet()
    test_siam_unet()
    test_unet3d()
    # delete test folder
    delete_folder_with_retry(folder)
    print("*" * 20 + " \nTests completed successfully")



import torch
from torch import nn as nn


class BCELoss2d(nn.Module):
    def __init__(self, weight=None, size_average=True):
        super(BCELoss2d, self).__init__()
        self.bce_loss = nn.BCEWithLogitsLoss(weight=weight, reduction='mean' if size_average else 'sum')

    def forward(self, logits, targets):
        return self.bce_loss(logits, targets)


class SoftDiceLoss(nn.Module):
    def __init__(self, smooth=1.0):
        super(SoftDiceLoss, self).__init__()
        self.smooth = smooth

    def forward(self, logits, targets):
        probs = torch.sigmoid(logits)
        num = targets.size(0)
        m1 = probs.view(num, -1)
        m2 = targets.view(num, -1)
        intersection = (m1 * m2).sum(1)
        score = 2. * (intersection + self.smooth) / (m1.sum(1) + m2.sum(1) + self.smooth)
        return 1 - score.mean()


class BCEDiceLoss(nn.Module):
    def __init__(self, alpha, beta):
        super(BCEDiceLoss, self).__init__()
        self.bce = BCELoss2d()
        self.dice = SoftDiceLoss()
        self.alpha = alpha
        self.beta = beta

    def forward(self, logits, targets):
        return self.alpha * self.bce(logits, targets) + self.beta * self.dice(logits, targets)


class logcoshDiceLoss(nn.Module):
    def __init__(self):
        super(logcoshDiceLoss, self).__init__()
        self.dice = SoftDiceLoss()

    def forward(self, logits, targets):
        x = self.dice(logits, targets)
        return torch.log((torch.exp(x) + torch.exp(-x)) / 2)


class TverskyLoss(nn.Module):
    def __init__(self, alpha=0.5, beta=0.5, smooth=1):
        super(TverskyLoss, self).__init__()
        self.alpha = alpha
        self.beta = beta
        self.smooth = smooth

    def forward(self, inputs, targets):
        inputs = torch.sigmoid(inputs)
        inputs = inputs.view(-1)
        targets = targets.view(-1)

        # True Positives, False Positives & False Negatives
        TP = (inputs * targets).sum()
        FP = ((1 - targets) * inputs).sum()
        FN = (targets * (1 - inputs)).sum()

        Tversky = (TP + self.smooth) / (TP + self.alpha * FP + self.beta * FN + self.smooth)

        return 1 - Tversky


class logcoshTverskyLoss(nn.Module):
    def __init__(self, alpha=0.5, beta=0.5, smooth=1):
        super(logcoshTverskyLoss, self).__init__()
        self.alpha = alpha
        self.beta = beta
        self.smooth = smooth

    def forward(self, inputs, targets):
        inputs = torch.sigmoid(inputs)
        inputs = inputs.view(-1)
        targets = targets.view(-1)

        # True Positives, False Positives & False Negatives
        TP = (inputs * targets).sum()
        FP = ((1 - targets) * inputs).sum()
        FN = (targets * (1 - inputs)).sum()

        Tversky = (TP + self.smooth) / (TP + self.alpha * FP + self.beta * FN + self.smooth)

        return torch.log(torch.cosh(1 - Tversky))



import glob
import os
import shutil
from skimage import morphology
from albumentations import (
    ShiftScaleRotate, Blur, MultiplicativeNoise,
    RandomBrightnessContrast, Flip, Compose, RandomRotate90)

import numpy as np
import tifffile
import torch
from torch.utils.data import Dataset


class DataProcess(Dataset):
    """Class for training data generation (processing, splitting, augmentation)"""

    def __init__(self, source_dir, dim_out=(256, 256), aug_factor=10, data_path='../data/', in_channels=1,
                 out_channels=1, dilate_mask=0, dilate_kernel='disk', add_tile=0,
                 val_split=0.2, invert=False, skeletonize=False, clip_threshold=(0.2, 99.8), shiftscalerotate=(0, 0, 0),
                 noise_lims=(0.5, 1.2), brightness_contrast=(0.25, 0.25), blur_limit=(2, 7), rescale=None, create=True):
        """
        Create training data object for network training

        1) Create folder structure for training data
        2) Move and preprocess training images
        3) Split input images into tiles
        4) Augment training data
        5) Create object of PyTorch Dataset class for training

        Parameters
        ----------
        source_dir : Tuple[str, str]
            Path of training data [images, labels]. Images need to be tif files.
        dim_out : Tuple[int, int]
            Resize dimensions of images for training
        aug_factor : int
            Factor of image augmentation
        data_path : str
            Base path of temporary directories for training data
        in_channels : int
            Number of input channels of training data
        out_channels : int
            Number of output channels of training data
        dilate_mask
            Radius of binary dilation of masks [-2, -1, 0, 1, 2]
        dilate_kernel : str
            Dilation kernel ('disk' or 'square')
        add_tile : int
            Add additional tile for splitting images into tiles with more overlapping tiles
        val_split : float
            Validation split for training
        invert : bool
            If True, greyscale binary labels is inverted
        skeletonize : bool
            If True, binary labels are skeletonized
        clip_threshold : Tuple[float, float]
            Clip thresholds for intensity normalization of images
        shiftscalerotate : [float, float, float]
            Shift, scale and rotate image during augmentation
        noise_amp : float
            Amplitude of Gaussian noise for image augmentation
        brightness_contrast : Tuple[float, float]
            Adapt brightness and contrast of images during augmentation
        rescale : float, optional
            Rescale all images and labels by factor rescale
        create : bool, optional
            If False, existing data set in data_path is used
        """
        self.source_dir = source_dir
        self.create = create
        self.data_path = data_path
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.dim_out = dim_out
        self.skeletonize = skeletonize
        self.invert = invert
        self.clip_threshold = clip_threshold
        self.add_tile = add_tile
        self.aug_factor = aug_factor
        self.shiftscalerotate = shiftscalerotate
        self.brightness_contrast = brightness_contrast
        self.noise_lims = noise_lims
        self.dilate_mask = dilate_mask
        self.dilate_kernel = dilate_kernel
        self.blur_limit = blur_limit
        self.val_split = val_split
        self.mode = 'train'

        self.__make_dirs()
        if create:
            self.__move_and_edit()
            self.__merge_images()
            self.__split()
            if self.aug_factor is not None:
                self.__augment()

    def __make_dirs(self):
        self.image_path = self.data_path + '/image/'
        self.mask_path = self.data_path + '/mask/'
        self.merge_path = self.data_path + '/merge/'
        self.split_merge_path = self.data_path + '/split/merge/'
        self.split_image_path = self.data_path + '/split/image/'
        self.split_mask_path = self.data_path + '/split/mask/'
        self.aug_image_path = self.data_path + '/augmentation/aug_image/'
        self.aug_mask_path = self.data_path + '/augmentation/aug_mask/'

        # delete old files
        if self.create:
            try:
                shutil.rmtree(self.data_path)
            except:
                pass
        # make folders
        os.makedirs(self.image_path, exist_ok=True)
        os.makedirs(self.mask_path, exist_ok=True)
        os.makedirs(self.merge_path, exist_ok=True)
        os.makedirs(self.split_merge_path, exist_ok=True)
        os.makedirs(self.split_image_path, exist_ok=True)
        os.makedirs(self.split_mask_path, exist_ok=True)
        os.makedirs(self.aug_image_path, exist_ok=True)
        os.makedirs(self.aug_mask_path, exist_ok=True)

    def __move_and_edit(self):
        # create image data
        files_image = glob.glob(self.source_dir[0] + '*')
        for file_i in files_image:
            img_i = tifffile.imread(file_i).astype(np.float32)
            # clip and normalize (0,255)
            img_i = np.clip(img_i, a_min=np.nanpercentile(img_i, self.clip_threshold[0]),
                            a_max=np.percentile(img_i, self.clip_threshold[1]))
            img_i = (img_i - np.nanmin(img_i)) / (np.nanmax(img_i) - np.nanmin(img_i)) * 255
            img_i = img_i.astype('uint8')
            save_i = os.path.splitext(os.path.basename(file_i))[0]
            save_i = save_i.replace(' ', '_')
            if len(img_i.shape) == 2:
                img_i = np.expand_dims(img_i, 0)
            tifffile.imwrite(self.image_path + save_i + '.tif', img_i)
        # create masks
        files_mask = glob.glob(self.source_dir[1] + '*')
        print('%s files found' % len(files_mask))
        for file_i in files_mask:
            mask_i = tifffile.imread(file_i)
            if len(mask_i.shape) == 2:
                mask_i = np.expand_dims(mask_i, 0)
            for j, ch_j in enumerate(mask_i):  # iterate channels
                if self.skeletonize:
                    ch_j = ch_j > 1
                    ch_j = morphology.skeletonize(ch_j) * 255
                if self.dilate_kernel == 'disk':
                    kernel = morphology.disk
                elif self.dilate_kernel == 'square':
                    kernel = morphology.square
                else:
                    raise ValueError(f'Dilate kernel {self.dilate_kernel} unknown!')
                if self.dilate_mask > 0 and self.dilate_mask:
                    ch_j = morphology.erosion(ch_j, kernel(self.dilate_mask))
                elif self.dilate_mask < 0:
                    ch_j = morphology.dilation(ch_j, kernel(-self.dilate_mask))
                if self.invert:
                    ch_j = 255 - ch_j
                mask_i[j] = ch_j
            mask_i = mask_i.astype('uint8')
            save_i = os.path.splitext(os.path.basename(file_i))[0]
            save_i = save_i.replace(' ', '_')
            tifffile.imwrite(self.mask_path + save_i + '.tif', mask_i.astype('int8'))

    def __merge_images(self):
        self.mask_files = glob.glob(self.data_path + '/mask/*.tif')
        self.image_files = glob.glob(self.data_path + '/image/*.tif')

        if len(self.mask_files) != len(self.image_files):
            raise ValueError('Number of ground truth does not match number of image stacks')

        for i, file_i in enumerate(self.mask_files):
            basename_i = os.path.basename(file_i)
            mask_i = tifffile.imread(self.data_path + '/mask/' + basename_i)
            img_i = tifffile.imread(self.data_path + '/image/' + basename_i)
            merge = np.zeros((mask_i.shape[1], mask_i.shape[2], mask_i.shape[0]+img_i.shape[0]))
            merge[:, :, :mask_i.shape[0]] = np.moveaxis(mask_i, 0, 2)
            merge[:, :, mask_i.shape[0]:] = np.moveaxis(img_i, 0, 2)
            merge = merge.astype('uint8')
            tifffile.imwrite(self.merge_path + str(i) + '.tif', merge)

    def __split(self):
        self.merges = glob.glob(self.merge_path + '*.tif')
        n = 0
        for i in range(len(self.merges)):
            merge = tifffile.imread(self.merge_path + str(i) + '.tif')
            dim_in = merge.shape
            # padding if dim_in < dim_out
            x_gap = max(0, self.dim_out[0] - dim_in[0])
            y_gap = max(0, self.dim_out[1] - dim_in[1])
            merge = np.pad(merge, ((0, x_gap), (0, y_gap), (0, 0)), 'reflect')
            # number of patches in x and y
            dim_in = merge.shape
            N_x = int(np.ceil(dim_in[0] / self.dim_out[0]))
            N_y = int(np.ceil(dim_in[1] / self.dim_out[1]))
            N_x += self.add_tile if N_x > 1 else 0
            N_y += self.add_tile if N_y > 1 else 0
            # starting indices of patches
            X_start = np.linspace(0, dim_in[0] - self.dim_out[0], N_x).astype('int16')
            Y_start = np.linspace(0, dim_in[1] - self.dim_out[1], N_y).astype('int16')
            for j in range(N_x):
                for k in range(N_y):
                    patch_ij = merge[X_start[j]:X_start[j] + self.dim_out[0], Y_start[k]:Y_start[k] + self.dim_out[1],
                               :]
                    mask_ij = patch_ij[:, :, :self.out_channels]
                    image_ij = patch_ij[:, :, self.out_channels:]
                    tifffile.imwrite(self.split_merge_path + f'{n}.tif', patch_ij)
                    tifffile.imwrite(self.split_mask_path + f'{n}.tif', mask_ij)
                    tifffile.imwrite(self.split_image_path + f'{n}.tif', image_ij)
                    n += 1

    def __augment(self, p=0.8):
        aug_pipeline = Compose(transforms=[
            Flip(),
            RandomRotate90(p=1.0),
            ShiftScaleRotate(self.shiftscalerotate[0], self.shiftscalerotate[1], self.shiftscalerotate[2]),
            RandomBrightnessContrast(brightness_limit=self.brightness_contrast[0],
                                     contrast_limit=self.brightness_contrast[1], p=0.5),
            Blur(blur_limit=self.blur_limit, always_apply=False, p=0.2),
            MultiplicativeNoise(multiplier=self.noise_lims, elementwise=True, p=0.3),
        ],
            p=p)

        patches_image = glob.glob(self.split_image_path + '*.tif')
        patches_mask = glob.glob(self.split_mask_path + '*.tif')
        n_patches = len(patches_image)
        k = 0
        for i in range(n_patches):
            image_i = tifffile.imread(patches_image[i])
            mask_i = tifffile.imread(patches_mask[i])

            data_i = {'image': image_i, 'mask': mask_i}
            data_aug_i = np.asarray([aug_pipeline(**data_i) for _ in range(self.aug_factor)])
            imgs_aug_i = np.asarray([data_aug_i[j]['image'] for j in range(self.aug_factor)])
            masks_aug_i = np.asarray([data_aug_i[j]['mask'] for j in range(self.aug_factor)])

            for j in range(self.aug_factor):
                tifffile.imwrite(self.aug_image_path + f'{k}.tif', np.moveaxis(imgs_aug_i[j], 2, 0))
                tifffile.imwrite(self.aug_mask_path + f'{k}.tif', np.moveaxis(masks_aug_i[j], 2, 0))
                k += 1
        print(f'Number of training images: {k}')

    def __len__(self):
        if self.aug_factor is not None:
            return len(os.listdir(self.aug_image_path))
        else:
            return len(os.listdir(self.split_image_path))

    def __getitem__(self, idx):
        imgname = str(idx) + '.tif'
        midname = os.path.basename(imgname)
        if self.aug_factor is not None:
            image_0 = tifffile.imread(self.aug_image_path + midname).astype('float32') / 255
            mask_0 = tifffile.imread(self.aug_mask_path + midname).astype('float32') / 255
        else:
            image_0 = tifffile.imread(self.split_image_path + midname).astype('float32') / 255
            mask_0 = tifffile.imread(self.split_image_path + midname).astype('float32') / 255
        image = torch.from_numpy(image_0)
        mask = torch.from_numpy(mask_0)
        del image_0, mask_0
        sample = {'image': image, 'mask': mask}
        return sample



import cv2 # type: ignore
import numpy as np
from util import write_info_file

def average_tifs(inputs : list, output):
	"""Averages a list of single channel tiff movies

	Args:
		inputs (list): A list of paths (str) to tiff movies
		output (str): Location to output the tiff file

	Raises:
		IOError: If one of the input tiff movies is not found
	"""
	tifs = []
	for tiff_movie in inputs:
		retval, imgs = cv2.imreadmulti(tiff_movie)
		if not retval:
			print(f'Error: {tiff_movie} not found.')
			raise IOError   # tiff file not found
		tifs.append(imgs)
	print('Averaging', len(tifs), 'images')
	out_tifs = [] # list (frames) of numpy arrays that are of size e.g. 1160*2300
	for i in range(len(tifs[0])):
		out_tifs.append(np.zeros(tifs[0][0].shape, dtype=np.int16))

	# add each frame from the input to the output
	for tif_idx, tif in enumerate(tifs):
		for frame_idx, frame in enumerate(tif):
			# add each frame to the corresponding frame in out_tif
			out_tifs[frame_idx] += frame

	write_info_file(f'{output}.info.txt', f'Operation: Average\nOutfile: {output}\nInput files:{inputs}')
	final_output_tif = []
	# average the frames in the output
	for frame in out_tifs:
		final_output_tif.append((frame / len(tifs)).astype('uint8'))
	
	cv2.imwritemulti(output, final_output_tif)


if __name__ == '__main__':
	pass


import cv2
import numpy as np
import matplotlib.pyplot as plt
import math

def create_pixel_value_histogram(input_tifs, frames_per_hist=100, bin_width=8):
	"""Creates a histogram for the pixel values in a tif file

	Args:
		input_tifs (str path to .tif files): input
		frames_per_hist (int, optional): number of frames to plot a histogram. Defaults to 100.
		bin_width (int, optional): bin width on the outpuut tif plot. Should be set to a value divisible by 256. Defaults to 8.

	Raises:
		IOError: An IOError is raised when the input tif file cannot be found
	"""
	retval, imgs = cv2.imreadmulti(input_tifs[0])
	fig, axs = plt.subplots(math.ceil(len(imgs)/frames_per_hist), len(input_tifs), figsize=(5, 20))
	fig.suptitle([input_tif.split('/')[-1] for input_tif in input_tifs])
	for input_tif_idx, input_tif in enumerate(input_tifs):
		if input_tif_idx != 0:
			retval, imgs = cv2.imreadmulti(input_tif)
		if not retval:
			print(f'Error: {input_tif} not found.')
			raise IOError   # tiff file not found		
		axis_idx = 0
		for idx, img in enumerate(imgs):
			if idx % frames_per_hist == 0:
				img_val = img.flatten()
				axs[axis_idx, input_tif_idx].hist(img_val, bins=range(0, 256, bin_width))
				axs[axis_idx, input_tif_idx].set(title=f'frame={idx}')
				axs[axis_idx, input_tif_idx].set_xlim([0, 255])
				axs[axis_idx, input_tif_idx].set_yscale('log')
				axis_idx += 1
	plt.show()


def main():
	# tif_name = constants.TIFS_NAMES[3]
	# create_pixel_value_histogram([f"/home/longyuxi/Documents/mount/predicted_out/old_self_model_predictions/predicted_{tif_name}_with_cosh_dice.tif", f"/home/longyuxi/Documents/mount/predicted_out/predicted_newly_drawn_{tif_name}_epoch_500.tif"], bin_width=2)
	pass



if __name__ == '__main__':
	main()


import numpy as np
import tifffile
from PIL import Image
from torch import nn
import torch


def save_as_tif(imgs, filename, normalize=False):
    """
    Save numpy array as tif file

    Parameters
    ----------
    imgs : np.array
        Data array
    filename : str
        Filepath to save result
    normalize : bool
        If true, data is normalized [0, 255]
    """
    if normalize:
        imgs = imgs.astype('float32')
        imgs = imgs - np.nanmin(imgs)
        imgs /= np.nanmax(imgs)
        imgs *= 255
    imgs = imgs.astype('uint8')
    tifffile.imwrite(filename, imgs)


def png_to_grayscale_tiff(png_filename, tiff_filename):
    """
    Read a PNG file, convert it to grayscale, normalize the image data, and save it as a TIFF file.

    Parameters
    ----------
    png_filename : str
        The filename of the PNG file to read.
    tiff_filename : str
        The filename of the TIFF file to save.

    Returns
    -------
    None
    """
    # Read the PNG file
    image = Image.open(png_filename)

    # Convert the image to grayscale
    grayscale_image = image.convert("L")

    # Normalize the image data by its maximum value
    normalized_image = np.array(grayscale_image) / np.max(grayscale_image) * 255

    # Convert the normalized image to np.uint8 data type
    normalized_image = normalized_image.astype(np.uint8)

    # Save the normalized image as a TIFF file
    tifffile.imwrite(tiff_filename, normalized_image)


def get_device(print_device=False):
    """
    Determines the most suitable device (CUDA, MPS, or CPU) for PyTorch operations.

    Returns:
    - A torch.device object representing the selected device.
    """
    if torch.backends.cuda.is_built():
        device = torch.device('cuda:0')
    elif torch.backends.mps.is_built():  # only for Apple M1/M2/...
        device = torch.device('mps')
    else:
        device = torch.device('cpu')
        print("Warning: No CUDA or MPS device found. Calculations will run on the CPU, "
              "which might be slower.")
    if print_device:
        print(f"Using device: {device}")
    return device


def init_weights(m):
    if isinstance(m, nn.Conv2d):
        nn.init.kaiming_normal_(m.weight, nonlinearity='leaky_relu')  # He initialization



import torch
from torch import nn


class BabyUnet(nn.Module):
    def __init__(self, n_filter=4):
        """
        Neural network for semantic image segmentation U-Net (PyTorch), with only three max-pooling layers
        Reference:  Falk, T. et al. U-Net: deep learning for cell counting, detection, and morphometry. Nat Methods 16,
        67–70 (2019).

        Parameters
        ----------
        n_filter : int
            Number of convolutional filters (commonly 2**n)
        """
        super().__init__()
        # encode
        self.encode1 = self.conv(1, n_filter)
        self.encode2 = self.conv(n_filter, n_filter)
        self.maxpool1 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.encode3 = self.conv(n_filter, 2 * n_filter)
        self.encode4 = self.conv(2 * n_filter, 2 * n_filter)
        self.maxpool2 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.encode5 = self.conv(2 * n_filter, 4 * n_filter)
        self.encode6 = self.conv(4 * n_filter, 4 * n_filter)
        self.maxpool3 = nn.MaxPool2d(kernel_size=2, stride=2)

        # middle
        self.middle_conv1 = self.conv(4 * n_filter, 8 * n_filter)
        self.middle_conv2 = self.conv(8 * n_filter, 8 * n_filter, dropout=0.5)

        # decode
        self.up1 = nn.ConvTranspose2d(8 * n_filter, 4 * n_filter, kernel_size=2, stride=2)
        self.decode1 = self.conv(8 * n_filter, 4 * n_filter)
        self.decode2 = self.conv(4 * n_filter, 4 * n_filter)
        self.up2 = nn.ConvTranspose2d(4 * n_filter, 2 * n_filter, kernel_size=2, stride=2)
        self.decode3 = self.conv(4 * n_filter, 2 * n_filter)
        self.decode4 = self.conv(2 * n_filter, 2 * n_filter)
        self.up3 = nn.ConvTranspose2d(2 * n_filter, 1 * n_filter, kernel_size=2, stride=2)
        self.decode5 = self.conv(2 * n_filter, 1 * n_filter)
        self.decode6 = self.conv(1 * n_filter, 1 * n_filter)
        self.decode7 = self.conv(1 * n_filter, 1)
        self.final = nn.Sequential(
            nn.Conv2d(1, 1, kernel_size=1, padding=0),
        )

    def conv(self, in_channels, out_channels, kernel_size=3, dropout=0.0):
        block = nn.Sequential(
            nn.Conv2d(kernel_size=kernel_size, in_channels=in_channels, out_channels=out_channels, padding=1),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(),
            nn.Dropout2d(dropout)
        )
        return block

    def concat(self, x1, x2):
        if x1.shape == x2.shape:
            return torch.cat((x1, x2), 1)
        else:
            print(x1.shape, x2.shape)
            raise ValueError('concatenation failed: wrong dimensions')

    def forward(self, x):
        e1 = self.encode1(x)
        e2 = self.encode2(e1)
        m1 = self.maxpool1(e2)
        e3 = self.encode3(m1)
        e4 = self.encode4(e3)
        m2 = self.maxpool2(e4)
        e5 = self.encode5(m2)
        e6 = self.encode6(e5)
        m3 = self.maxpool3(e6)

        mid1 = self.middle_conv1(m3)
        mid2 = self.middle_conv2(mid1)

        u1 = self.up1(mid2)
        c1 = self.concat(u1, e5)
        d1 = self.decode1(c1)
        d2 = self.decode2(d1)
        u2 = self.up2(d2)
        c2 = self.concat(u2, e3)
        d3 = self.decode3(c2)
        d4 = self.decode4(d3)
        u3 = self.up3(d4)
        c3 = self.concat(u3, e1)
        d5 = self.decode5(c3)
        d6 = self.decode6(d5)
        d9 = self.decode7(d6)
        logits = self.final(d9)
        return torch.sigmoid(logits), logits



import cv2 # type: ignore
import numpy as np
import time
import tifffile
from tifffile import TiffFile
import os

def find_frame_of_image(query_image, search_space=[], save_machine_readable_output=True, machine_readable_output_filename='search_result_mr.txt'):
    """Finds the frame number of query_image within search_space. 
    
    This function will move on to the next candidate in search_space if the dimensions of query_image doesn't match a certain candidate. If the dimensions of query_image and candidate match, this function first attempts to find an exact match, but resorts to finding frame with the least mean squared error (MSE) if no exact match is found.

    If save_machine_readable_output is set to True, this function will write a tab-separated file containing the names (not absolute paths) of the query image (label), the candidate image with the best match, and the frame number (0-indexed) that has the best match.

    This function is best used with the function utilize_search_result.generate_siam_unet_input_imgs().

    Args:
        query_image (str): path to query tiff image, must be single frame
        search_space (list of str, optional): The list of images over which to search query_image. Defaults to constants.TIFS_NAMES.
    """

    query = cv2.imread(query_image, cv2.IMREAD_GRAYSCALE)
    for candidate in search_space:
        tif_key = TiffFile(candidate)
        # tif_len = len(tif_key.pages)

        # check if dimensions are the same
        if not tif_key.pages[0].shape == query.shape:
            print(f'Shape of query {query.shape} differs with shape of {candidate} {tif_key.pages[0].shape}')
            continue

        # if yes, progress
        mses = []
        frame_nbr = 0
        imgs = frame_generator(candidate)
        for img in imgs:
            # print(frame_nbr)
            if np.array_equal(img, query):
                print(f"\n\n!!!!Found exact match in frame {frame_nbr} of {candidate}.\n\n")
                continue
            mses.append(mse(query, img))
            frame_nbr += 1
        
        # if there is not an exact match, resort to the best MSE solution
        print(f'No exact match was found in {candidate}. The closest matching frame was {mses.index(min(mses))} with MSE of {min(mses)}')


        if save_machine_readable_output:
            # save information directly if:
            #   1. frame numbers match
            #   2. MSE < 1000
            if mses.index(min(mses)) == int(os.path.basename(query_image).split(".")[0]):
                if min(mses) < 1000:
                    with open(machine_readable_output_filename, 'a') as o:
                        o.write(f'{os.path.basename(query_image)}\t{os.path.basename(candidate)}\t{mses.index(min(mses))}\n')

def frame_generator(tiff_movie):
    """Returns a generator for each individual frame of the movie. Should function in the same way as imgs in the expression "for img in imgs".

    Args:
        tiff_movie (str): path to tiff movie
    """
    tif_key = tifffile.TiffFile(tiff_movie)
    tif_len = len(tif_key.pages)
    for i in range(tif_len):
        yield tifffile.imread(tiff_movie, key=i)

def mse(imageA, imageB):
    # the 'Mean Squared Error' between the two images is the
    # sum of the squared difference between the two images;
    # NOTE: the two images must have the same dimension
    err = np.sum((imageA.astype("float") - imageB.astype("float")) ** 2)
    err /= float(imageA.shape[0] * imageA.shape[1])
    
    # return the MSE, the lower the error, the more "similar"
    # the two images are
    return err


if __name__ == '__main__':
    # find_frame_of_image(f'{constants.LOCAL_OUT_DIR}/21B25_shgGFP_kin_1_Pos0_frame_10.tif', [f'{constants.LOCAL_MOVIE_DIR}/{constants.TIFS_NAMES[1]}.tif'])

    # the missing search space was 21D16_shgGFPkin_Pos7.tif


    # # successful at finding the location to the amnioserosa images
    # file_names = ['105.tif','111.tif','120.tif','121.tif','146.tif','165.tif','166.tif','167.tif','1.tif','212.tif','224.tif','231.tif','253.tif','268.tif','2.tif','305.tif','314.tif','328.tif','331.tif','335.tif','369.tif','372.tif','3.tif','413.tif','439.tif','480.tif','496.tif','502.tif','510.tif','522.tif','5.tif','625.tif','646.tif','681.tif','69.tif','700.tif','707.tif','732.tif','778.tif','792.tif','7.tif','822.tif','837.tif','83.tif','840.tif']

    # # file_names = ['105.tif']
    # for file_name in file_names:
    #     print(file_name)
    #     find_frame_of_image(f'{constants.RAZER_LOCAL_BASE_DIR}/training_data/training_data/yokogawa/amnioserosa/image/{file_name}', search_space=constants.RAZER_LARGE_SEARCH_SPACE)
    pass


from typing import Union

import numpy as np
import tifffile
import torch

from ..progress import ProgressNotifier
from .unet import Unet
from .unet_v0 import Unet_v0
from .attention_unet import AttentionUnet
from ..utils import save_as_tif, get_device


class Predict:
    """Class for prediction of movies and images with U-Net"""

    def __init__(self, imgs, result_name, model_params, network='Unet', resize_dim=(512, 512),
                 invert=False, normalization_mode='single', clip_threshold=(0., 99.8), add_tile=0,
                 normalize_result=False, show_progress=True, device: Union[torch.device, str] = 'auto',
                 progress_notifier: ProgressNotifier = ProgressNotifier.progress_notifier_tqdm()):
        """
        Prediction of tif files with standard 2D U-Net

        1) Loading file and preprocess (normalization)
        2) Resizing of images into patches with resize_dim
        3) Prediction with U-Net
        4) Stitching of predicted patches and averaging of overlapping regions

        Parameters
        ----------
        imgs : ndarray, string
            images to predict, if string, attempt to load from tif file
        result_name : str
            path for result
        model_params : str
            path of u-net parameters (.pth file)
        network
            Network class
        resize_dim
            Image dimensions for resizing for prediction
        invert : bool
            Invert greyscale of image(s) before prediction
        normalization_mode : str
            Mode for intensity normalization for 3D stacks prior to prediction ('single': each image individually,
            'all': based on histogram of full stack, 'first': based on histogram of first image in stack)
        clip_threshold : Tuple[float, float]
            Clip threshold for image intensity before prediction
        add_tile : int, optional
            Add additional tiles for splitting large images to increase overlap
        normalize_result : bool
            If true, results are normalized to [0, 255]
        show_progress : bool
            Whether to show progress bar.
        device : torch.device or str, optional
            Device to run the pytorch model on, defaults to 'auto', which selects CUDA or MPS if available.
        progress_notifier:
            Wrapper to show tqdm progress notifier in gui
        """
        if device == 'auto':
            self.device = get_device()
        else:
            self.device = torch.device(device)

        if isinstance(imgs, str):
            imgs = tifffile.imread(imgs)

        self.resize_dim = resize_dim
        self.add_tile = add_tile
        self.normalize_result = normalize_result
        self.invert = invert
        self.normalization_mode = normalization_mode
        self.clip_threshold = clip_threshold
        self.result_name = result_name
        self.show_progress = show_progress

        # read, preprocess and split data
        imgs = self.__reshape_data(imgs)
        imgs = self.__preprocess(imgs)
        patches = self.__split(imgs)
        del imgs

        # load model and predict data
        self.model_params = torch.load(model_params, map_location=self.device)
        if network is None:
            if 'network' in self.model_params.keys():
                network = self.model_params['network']
            else:
                raise ValueError('network is not defined')
        if network == 'Unet':
            network = Unet
        elif network == 'AttentionUnet':
            network = AttentionUnet
        elif network == 'Unet_v0':
            network = Unet_v0
            if 'in_channels' not in self.model_params.keys():
                self.model_params['in_channels'] = 1
                self.model_params['out_channels'] = 1
        self.model = network(n_filter=self.model_params['n_filter'], in_channels=self.model_params['in_channels'],
                             out_channels=self.model_params['out_channels']).to(self.device)
        self.model.load_state_dict(self.model_params['state_dict'])
        self.model.eval()
        result_patches = self.__predict(patches, progress_notifier)
        del patches, self.model

        # stitch patches (mean of overlapped regions)
        imgs_result = self.__stitch(result_patches)
        del result_patches, self.model_params

        # save as .tif file
        save_as_tif(imgs_result, self.result_name, normalize=normalize_result)
        del imgs_result
        if torch.cuda.is_available():
            torch.cuda.empty_cache()

    def __reshape_data(self, imgs):
        self.imgs_shape = imgs.shape
        if len(self.imgs_shape) == 2:  # if single image
            imgs = np.expand_dims(imgs, axis=0)
            self.imgs_shape = imgs.shape
        return imgs

    def __preprocess(self, imgs):
        if self.normalization_mode == 'single':
            for i, img in enumerate(imgs):
                img = np.clip(img, a_min=np.nanpercentile(img, self.clip_threshold[0]),
                              a_max=np.percentile(img, self.clip_threshold[1]))
                img = img - np.min(img)
                img = img / np.max(img) * 255
                if self.invert:
                    img = 255 - img
                imgs[i] = img
        elif self.normalization_mode == 'first':
            clip_threshold = (np.nanpercentile(imgs[0], self.clip_threshold[0]),
                              np.percentile(imgs[0], self.clip_threshold[1]))
            imgs = np.clip(imgs, clip_threshold[0], clip_threshold[1])
            imgs = imgs - np.min(imgs)
            imgs = imgs / np.max(imgs) * 255
            if self.invert:
                imgs = 255 - imgs
        elif self.normalization_mode == 'all':
            clip_threshold = (np.nanpercentile(imgs, self.clip_threshold[0]),
                              np.percentile(imgs, self.clip_threshold[1]))
            imgs = np.clip(imgs, clip_threshold[0], clip_threshold[1])
            imgs = imgs - np.min(imgs)
            imgs = imgs / np.max(imgs) * 255
            if self.invert:
                imgs = 255 - imgs
        else:
            raise ValueError(f'normalization_mode {self.normalization_mode} not valid!')
        return imgs

    def __split(self, imgs):
        # number of patches in x and y
        self.N_x = int(np.ceil(self.imgs_shape[1] / self.resize_dim[0])) + self.add_tile
        self.N_y = int(np.ceil(self.imgs_shape[2] / self.resize_dim[1])) + self.add_tile
        self.N_per_img = self.N_x * self.N_y
        self.N = self.N_x * self.N_y * self.imgs_shape[0]  # total number of patches

        # define array for prediction
        patches = np.zeros((self.N, 1, self.resize_dim[0], self.resize_dim[1]), dtype='uint8')

        # zero padding of image if imgs_shape < resize_dim
        if self.resize_dim[0] > self.imgs_shape[1]:  # for x
            imgs = np.pad(imgs, ((0, 0), (0, self.resize_dim[0] - self.imgs_shape[1]), (0, 0)),
                          'reflect')
        if self.resize_dim[1] > self.imgs_shape[2]:  # for y
            imgs = np.pad(imgs, ((0, 0), (0, 0), (0, self.resize_dim[1] - self.imgs_shape[2])),
                          'reflect')

        # starting indices of patches
        self.X_start = np.linspace(0, self.imgs_shape[1] - self.resize_dim[0], self.N_x).astype('uint16')
        self.Y_start = np.linspace(0, self.imgs_shape[2] - self.resize_dim[1], self.N_y).astype('uint16')

        # split in resize_dim
        n = 0
        for i, img in enumerate(imgs):
            for j in range(self.N_x):
                for k in range(self.N_y):
                    patches[n, 0, :, :] = img[self.X_start[j]:self.X_start[j] + self.resize_dim[0],
                                              self.Y_start[k]:self.Y_start[k] + self.resize_dim[1]]
                    n += 1
        return patches

    def __predict(self, patches, progress_notifier: ProgressNotifier = ProgressNotifier.progress_notifier_tqdm()):
        result_patches = np.zeros((patches.shape[0], self.model_params['out_channels'], *patches.shape[2:]),
                                  dtype='uint8')
        print('Predicting data ...') if self.show_progress else None
        with torch.no_grad():
            _progress_notifier = enumerate(progress_notifier.iterator(patches)) if self.show_progress else enumerate(
                patches)
            for i, patch_i in _progress_notifier:
                patch_i = torch.from_numpy(patch_i.astype('float32') / 255).to(self.device).view((1,
                                                                                             self.model_params[
                                                                                                 'in_channels'],
                                                                                             self.resize_dim[0],
                                                                                             self.resize_dim[1]))
                res_i, logits_i = self.model(patch_i)
                res_i = res_i.view(
                    (self.model_params['out_channels'], self.resize_dim[0], self.resize_dim[1])).cpu().numpy()
                result_patches[i] = (res_i * 255).astype('uint8')
                del patch_i, res_i
        return result_patches

    def __stitch(self, result_patches):
        # create array
        imgs_result = np.zeros(
            (self.imgs_shape[0], self.model_params['out_channels'], np.max((self.resize_dim[0], self.imgs_shape[1])),
             np.max((self.resize_dim[1], self.imgs_shape[2]))), dtype='uint8')
        for i in range(self.imgs_shape[0]):

            stack_result_i = np.zeros((self.N_per_img, self.model_params['out_channels'],
                                       np.max((self.resize_dim[0], self.imgs_shape[1])),
                                       np.max((self.resize_dim[1], self.imgs_shape[2]))), dtype='uint8') * np.nan

            n = 0
            for j in range(self.N_x):
                for k in range(self.N_y):
                    stack_result_i[n, :, self.X_start[j]:self.X_start[j] + self.resize_dim[0],
                    self.Y_start[k]:self.Y_start[k] + self.resize_dim[1]] = result_patches[i * self.N_per_img + n]
                    n += 1

            # average overlapping regions
            imgs_result[i] = np.nanmean(stack_result_i, axis=0)
            del stack_result_i

        # change to input size (if zero padding)
        imgs_result = imgs_result[:, :, :self.imgs_shape[1], :self.imgs_shape[2]]

        return np.squeeze(imgs_result)



import torch
from torch import nn
import torch.nn.functional as F
import logging


class Siam_UNet(nn.Module):
    def __init__(self, n_filter=32, mode='concat'):
        super().__init__()
        # mode for combining T-1 and T
        self.mode = mode
        # encode
        self.encode1 = self.conv(1, n_filter)
        self.encode2 = self.conv(n_filter, n_filter)
        self.maxpool1 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.encode3 = self.conv(n_filter, 2 * n_filter)
        self.encode4 = self.conv(2 * n_filter, 2 * n_filter)
        self.maxpool2 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.encode5 = self.conv(2 * n_filter, 4 * n_filter)
        self.encode6 = self.conv(4 * n_filter, 4 * n_filter)
        self.maxpool3 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.encode7 = self.conv(4 * n_filter, 8 * n_filter)
        self.encode8 = self.conv(8 * n_filter, 8 * n_filter)
        self.maxpool4 = nn.MaxPool2d(kernel_size=2, stride=2)

        # middle
        if mode == 'concat':
            self.conv_concat = self.conv(16 * n_filter, 8 * n_filter)
        self.middle_conv1 = self.conv(8 * n_filter, 16 * n_filter)
        self.middle_conv2 = self.conv(16 * n_filter, 16 * n_filter)

        # decode
        self.up1 = nn.ConvTranspose2d(16 * n_filter, 8 * n_filter, kernel_size=2, stride=2)
        self.decode1 = self.conv(16 * n_filter, 8 * n_filter)
        self.decode2 = self.conv(8 * n_filter, 8 * n_filter)
        self.up2 = nn.ConvTranspose2d(8 * n_filter, 4 * n_filter, kernel_size=2, stride=2)
        self.decode3 = self.conv(8 * n_filter, 4 * n_filter)
        self.decode4 = self.conv(4 * n_filter, 4 * n_filter)
        self.up3 = nn.ConvTranspose2d(4 * n_filter, 2 * n_filter, kernel_size=2, stride=2)
        self.decode5 = self.conv(4 * n_filter, 2 * n_filter)
        self.decode6 = self.conv(2 * n_filter, 2 * n_filter)
        self.up4 = nn.ConvTranspose2d(2 * n_filter, 1 * n_filter, kernel_size=2, stride=2)
        self.decode7 = self.conv(2 * n_filter, 1 * n_filter)
        self.decode8 = self.conv(1 * n_filter, 1 * n_filter)
        self.final = nn.Sequential(
            nn.Conv2d(n_filter, 1, kernel_size=1, padding=0),
        )

    def conv(self, in_channels, out_channels, kernel_size=3, dropout=0., dilation=1):
        layers = [
            nn.Conv2d(in_channels, out_channels, kernel_size, padding=dilation, dilation=dilation),
            nn.BatchNorm2d(out_channels),
            nn.LeakyReLU(negative_slope=0.1, inplace=True),  # Using LeakyReLU
            nn.Dropout2d(dropout)]
        return nn.Sequential(*layers)

    def concat(self, x1, x2):
        if x1.shape == x2.shape:
            return torch.cat((x1, x2), 1)
        else:
            # logging.critical(x1.shape, x2.shape)
            logging.critical(f"Shapes: {x1.shape}, {x2.shape}")
            raise ValueError('concatenation failed: wrong dimensions')

    def depthwise_xcorr(self, embed_curr, embed_prev):
        """depth-wise cross correlation"""
        batch = embed_prev.size(0)
        channel = embed_prev.size(1)
        embed_curr = embed_curr.view(1, batch * channel, embed_curr.size(2), embed_curr.size(3))
        embed_prev = embed_prev.view(batch * channel, 1, embed_prev.size(2), embed_prev.size(3))
        out = F.conv2d(embed_curr, embed_prev, groups=batch * channel, padding='same')
        out = out.view(batch, channel, out.size(2), out.size(3))
        return out

    def forward(self, x, prev_x):
        # top encoder (current frame)
        e1 = self.encode1(x)
        e2 = self.encode2(e1)
        m1 = self.maxpool1(e2)
        e3 = self.encode3(m1)
        e4 = self.encode4(e3)
        m2 = self.maxpool2(e4)
        e5 = self.encode5(m2)
        e6 = self.encode6(e5)
        m3 = self.maxpool2(e6)
        e7 = self.encode7(m3)
        e8 = self.encode8(e7)
        m4 = self.maxpool2(e8)

        # bottom encoder (previous frame)
        me1 = self.encode1(prev_x)
        me2 = self.encode2(me1)
        mm1 = self.maxpool1(me2)
        me3 = self.encode3(mm1)
        me4 = self.encode4(me3)
        mm2 = self.maxpool2(me4)
        me5 = self.encode5(mm2)
        me6 = self.encode6(me5)
        mm3 = self.maxpool2(me6)
        me7 = self.encode7(mm3)
        me8 = self.encode8(me7)
        mm4 = self.maxpool2(me8)

        if self.mode == 'corr':
            join = self.depthwise_xcorr(m4, mm4)
        elif self.mode == 'max':
            join = torch.maximum(m4, mm4)
        elif self.mode == 'concat':
            conc = self.concat(m4, mm4)
            join = self.conv_concat(conc)
        elif self.mode == 'control':
            join = m4
        else:
            raise NotImplementedError('Unknown mode: {}'.format(self.mode))

        # bottleneck
        mid1 = self.middle_conv1(join)
        mid2 = self.middle_conv2(mid1)

        # decoder
        u1 = self.up1(mid2)
        c1 = self.concat(u1, e8)
        d1 = self.decode1(c1)
        d2 = self.decode2(d1)
        u2 = self.up2(d2)
        c2 = self.concat(u2, e6)
        d3 = self.decode3(c2)
        d4 = self.decode4(d3)
        u3 = self.up3(d4)
        c3 = self.concat(u3, e4)
        d5 = self.decode5(c3)
        d6 = self.decode6(d5)
        u4 = self.up4(d6)
        c4 = self.concat(u4, e2)
        d7 = self.decode7(c4)
        d8 = self.decode8(d7)
        logits = self.final(d8)
        return torch.sigmoid(logits), logits



import torch
from torch import nn
import torch.nn.functional as F


class UNet3D(nn.Module):
    def __init__(self, in_channels=1, out_channels=1, n_filter=16, use_interpolation=False):
        """
        Neural network for time-consistent segmentation or volume segmentation,
        adapted from Li, X. et al. Real-time denoising enables high-sensitivity fluorescence time-lapse imaging
        beyond the shot-noise limit. Nat Biotechnol 41, 282–292 (2023).


        Parameters
        ----------
        n_filter : int
            Number of convolutional filters (commonly 16, 32, or 64)
        """
        super().__init__()
        self.use_interpolation = use_interpolation

        # encode
        self.encode1 = self.conv3D(in_channels=in_channels, out_channels=n_filter // 2)
        self.encode2 = self.conv3D(n_filter // 2, n_filter)
        self.maxpool1 = nn.MaxPool3d(kernel_size=2, stride=2)
        self.encode3 = self.conv3D(n_filter, n_filter)
        self.encode4 = self.conv3D(n_filter, 2 * n_filter)
        self.maxpool2 = nn.MaxPool3d(kernel_size=2, stride=2)
        self.encode5 = self.conv3D(2 * n_filter, 2 * n_filter)
        self.encode6 = self.conv3D(2 * n_filter, 4 * n_filter)
        self.maxpool3 = nn.MaxPool3d(kernel_size=2, stride=2)

        # middle
        self.middle_conv1 = self.conv3D(4 * n_filter, 4 * n_filter)
        self.middle_conv2 = self.conv3D(4 * n_filter, 8 * n_filter)

        # decode
        if not use_interpolation:
            self.up1 = nn.ConvTranspose3d(8 * n_filter, 8 * n_filter, kernel_size=2, stride=2)
            self.up2 = nn.ConvTranspose3d(4 * n_filter, 4 * n_filter, kernel_size=2, stride=2)
            self.up3 = nn.ConvTranspose3d(2 * n_filter, 2 * n_filter, kernel_size=2, stride=2)

        self.decode1 = self.conv3D(12 * n_filter, 4 * n_filter)
        self.decode2 = self.conv3D(4 * n_filter, 4 * n_filter)
        self.decode3 = self.conv3D(6 * n_filter, 2 * n_filter)
        self.decode4 = self.conv3D(2 * n_filter, 2 * n_filter)
        self.decode5 = self.conv3D(3 * n_filter, n_filter)
        self.decode6 = self.conv3D(n_filter, n_filter // 2)
        self.final = nn.Conv3d(n_filter // 2, out_channels=out_channels, kernel_size=1, padding=0)

    def conv3D(self, in_channels, out_channels, kernel_size=3, dropout=0., dilation=1):
        layers = [
            nn.Conv3d(in_channels, out_channels, kernel_size, padding=dilation, dilation=dilation),
            nn.BatchNorm3d(out_channels),
            nn.LeakyReLU(negative_slope=0.1, inplace=True),  # Using LeakyReLU
            nn.Dropout3d(dropout)]
        return nn.Sequential(*layers)

    def concat(self, x1, x2):
        return torch.cat((x1, x2), 1)

    def forward(self, x):
        e1 = self.encode1(x)
        e2 = self.encode2(e1)
        m1 = self.maxpool1(e2)
        e3 = self.encode3(m1)
        e4 = self.encode4(e3)
        m2 = self.maxpool2(e4)
        e5 = self.encode5(m2)
        e6 = self.encode6(e5)
        m3 = self.maxpool3(e6)

        mid1 = self.middle_conv1(m3)
        mid2 = self.middle_conv2(mid1)

        if self.use_interpolation:
            u1 = F.interpolate(mid2, scale_factor=2, mode='trilinear', align_corners=False)
        else:
            u1 = self.up1(mid2)
        c1 = self.concat(u1, e6)
        d1 = self.decode1(c1)
        d2 = self.decode2(d1)
        if self.use_interpolation:
            u2 = F.interpolate(d2, scale_factor=2, mode='trilinear', align_corners=False)
        else:
            u2 = self.up2(d2)
        c2 = self.concat(u2, e4)
        d3 = self.decode3(c2)
        d4 = self.decode4(d3)
        if self.use_interpolation:
            u3 = F.interpolate(d4, scale_factor=2, mode='trilinear', align_corners=False)
        else:
            u3 = self.up3(d4)
        c3 = self.concat(u3, e2)
        d5 = self.decode5(c3)
        d6 = self.decode6(d5)
        logits = self.final(d6)
        return torch.sigmoid(logits), logits
